/*
 * smbsend.c: send interactive SMB (aka "winpopup") message
 *
 * Copyright 2003 by Peter Samuelson
 * Copyright 1994-1998 by Andrew Tridgell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: NONE */

#include "includes.h"
#include "libsamba.h"
#include "libnmb.h"
#include "libsmb.h"
#include "pathnames.h"

/* Observed message limits:
 * NT4:		981 bytes, displays empty message if larger
 * Win2k SP3:	4192 bytes, truncates message if larger
 * XP:		4029 bytes, ignores message if larger
 */
#define SMB_MSG_MAX_LEN 4029
#define SMB_MSG_MAX_SEG 127

extern pstring global_scope;
extern pstring global_myname;
extern BOOL in_client;

static BOOL get_name_03(struct nmb_name *n, const char *name, struct in_addr *ip)
{
	char namebuf[16];

	if (!isdigit(name[0])) {
		zero_ip(ip);
		make_nmb_name(n, name, 3);
		return True;
	}
	*ip = *interpret_addr2(name);
	if (name_status_find(3, *ip, namebuf)) {
		make_nmb_name(n, namebuf, 3);
		return True;
	}
	return False;
}


static void usage (char *progname, int exitval)
{
	fprintf(exitval ? stderr : stdout,
		"Usage: %s -h: this help text\n"
		"       %s [options] {host|user}...\n"
		"--- basic options ---\n"
		"  -F from        set the message sender's name (default: local NetBIOS name)\n"
		"  -T to          set the message receiver's name (default: remote NetBIOS name)\n"
		"  -m message     enter message on command line (default: standard input)\n"
		"  -f file        read message from file\n"
		"  {host|user}... at least one name (NetBIOS, DNS, IP addr)\n"
		"--- advanced options ---\n"
		"  -s smb.conf    override the location of smb.conf\n"
		"  -R order       override NetBIOS name resolve order\n"
		"  -d spec        specify a debug level spec\n"
		"  -i scope       set NetBIOS scope\n"
		"  -n myname      set NetBIOS sender name\n"
		"  -t termtype    set the terminal I/O code (sjis,euc,jis7,jis8,junet,hex)\n",
		progname, progname);
	exit(exitval);
}


int main (int argc, char *argv[])
{
	int c, msgfd;
	int failed = 0;
	char *progname;
	char *fromname = NULL, *toname = NULL;
	struct cli_state *cli = NULL;
	char *myname = NULL;
	char *msgfile = NULL;
	char *msg = NULL;
	int msglen = 0;
	int port = SMB_PORT;
	char *term_code = NULL;
	char *name_res_order = NULL;
	char *debug_spec = NULL;

	debug_parse_levels("1");
	dbf = stderr;

	progname = strrchr(argv[0], '/');
	if (!progname++)
		progname = argv[0];

	while ((c = getopt(argc, argv, "hd:F:f:i:m:p:R:s:T:t:")) != EOF) {
		switch(c) {
		case 'h':
			usage(progname, 0);
		case 'd':
			debug_spec = optarg;
			break;
		case 'F':
			fromname = optarg;
			break;
		case 'f':
			msgfile = optarg;
			break;
		case 'i':
			pstrcpy(global_scope, optarg);
			strupper(global_scope);
			break;
		case 'm':
			msg = optarg;
			msglen = strlen(msg);
			break;
		case 'n':
			myname = optarg;
			break;
		case 'p':
			port = atoi(optarg);
			if (port <= 0 || port > 65535)
				usage(progname, 1);
			break;
		case 'R':
			name_res_order = optarg;
			break;
		case 's':
			pathname_smb_conf = optarg;
			break;
		case 'T':
			toname = optarg;
			break;
		case 't':
			term_code = optarg;
			break;
		}
	}

	charset_initialise();

	in_client = True;   /* for lp_load */

	if (!lp_load(pathname_smb_conf, True, False, False))
		DEBUG(0,("Warning: load of %s failed\n", pathname_smb_conf));

	if (debug_spec)
		debug_parse_levels(debug_spec);

	codepage_initialise(lp_client_code_page());

	load_interfaces();

	if (myname) {
		pstrcpy(global_myname, myname);
		strupper(global_myname);
	}
	else
		get_myname(global_myname);

	if (!fromname)
		fromname = global_myname;

	if (name_res_order)
		lp_set_name_resolve_order(name_res_order);

	if (term_code)
		interpret_coding_system(term_code);

	argv += optind;
	argc -= optind;
	if (!argc)
		usage(progname, 1);

	/* read message from file or stdin */
	if (!msg) {
		static char buf[SMB_MSG_MAX_LEN];
		
		msgfd = STDIN_FILENO;
		if (msgfile) {
			msgfd = open(msgfile, O_RDONLY);
			if (msgfd < 0) {
				fprintf(stderr,
					"Cannot open input file %s\n", msgfile);
				exit(1);
			}
		}
		else if (isatty(msgfd))
			fprintf(stderr, "Type your message:\n");

		for (msglen = 0; msglen < SMB_MSG_MAX_LEN; ) {
			ssize_t r = read(msgfd, buf + msglen,
					 SMB_MSG_MAX_LEN - msglen);
			if (r <= 0)
				break;
			msglen += r;
		}
		if (msgfile)
			close(msgfd);
		msg = buf;
	}

	for (; argc; argv++,argc--) {
		struct nmb_name n_cli, n_srv;
		struct in_addr dest_ip;
		int msggrp;
		int i;
		char *to_name;

		make_nmb_name(&n_cli, global_myname, 0);
		if (!get_name_03(&n_srv, argv[0], &dest_ip)) {
			fprintf(stderr,
				"cannot find <03> name for %s\n", argv[0]);
			goto fail;
		}

		if (toname)
			to_name = toname;
		else
			to_name = n_srv.name;

		cli = cli_initialise(cli);
		if (!cli) {
			DEBUG(0,("cli_initialise failed for %s\n", argv[0]));
			goto fail;
		}
		if (!cli_connect(cli, argv[0], 3, &dest_ip)) {
			DEBUG(0,("cli_connect failed for %s\n", argv[0]));
			goto fail;
		}
		if (!cli_session_request(cli, &n_cli, &n_srv)) {
			DEBUG(0,("cli_session_request failed for %s\n", argv[0]));
			goto fail;
		}
		if (!cli_message_start(cli, to_name, fromname, &msggrp)) {
			DEBUG(0,("cli_message_start failed for %s\n",argv[0]));
			goto fail;
		}
		for (i=0; i<msglen; ) {
			int len = MIN(msglen-i, SMB_MSG_MAX_SEG);
			if (!cli_message_text(cli, msg + i, len, msggrp)) {
				failed = 1;
				break;
			}
			i += len;
		}
		cli_message_end(cli, msggrp);
		DEBUG(1,("Sent %d bytes to %s\n", i, argv[0]));
		continue;
	fail:
		failed = 1;
	}

	if (cli)
		cli_shutdown(cli);
	exit(failed);
}
