
/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2000,
 *  Copyright (C) Jeremy Allison               1998-2000.
 *  Copyright (C) Elrond                            2000.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "nterr.h"
#include "sids.h"
#include "others.h"


#define RPC_SERVER_STUB_DECLARE(name) \
	static BOOL name(PRS_HANDLE *hnd, prs_struct *data, prs_struct *rdata)



/***************************************************************************
make_reply_lookup_names
 ***************************************************************************/
static void make_reply_lookup_names(LSA_R_LOOKUP_NAMES * r_l,
				    DOM_R_REF * ref, uint32 num_entries,
				    DOM_RID2 * rid2, uint32 mapped_count,
				    uint32 status)
{
	r_l->ptr_dom_ref = (ref != NULL ? 1 : 0);
	r_l->dom_ref = ref;

	if (rid2 == NULL)
		num_entries = 0;

	r_l->num_entries = num_entries;
	r_l->ptr_entries = (rid2 != NULL ? 1 : 0);
	r_l->num_entries2 = num_entries;
	r_l->dom_rid = rid2;

	r_l->mapped_count = mapped_count;

	r_l->status = status;
}

/***************************************************************************
make_reply_lookup_sids
 ***************************************************************************/
static void make_reply_lookup_sids(LSA_R_LOOKUP_SIDS * r_l,
				   DOM_R_REF * ref,
				   LSA_TRANS_NAME_ENUM * names,
				   uint32 mapped_count, uint32 status)
{
	r_l->ptr_dom_ref = 1;
	r_l->dom_ref = ref;
	r_l->names = names;
	r_l->mapped_count = mapped_count;
	r_l->status = status;
}

/***************************************************************************
lsa_reply_lookup_sids
 ***************************************************************************/
static BOOL lsa_reply_lookup_sids(LSA_Q_LOOKUP_SIDS * q_l, prs_struct *rdata)
{
	LSA_R_LOOKUP_SIDS r_l;
	DOM_R_REF ref;
	LSA_TRANS_NAME_ENUM names;
	uint32 mapped_count = 0;
	DOM_SID2 *sid = q_l->sids.sid;
	int num_entries = q_l->sids.num_entries;
	uint32 status;

	ZERO_STRUCT(r_l);
	ZERO_STRUCT(ref);
	ZERO_STRUCT(names);

	/* set up the LSA Lookup SIDs response */
	status = _lsa_lookup_sids(&q_l->pol,
				  num_entries, sid, &q_l->level,
				  &ref, &names, &mapped_count);
	make_reply_lookup_sids(&r_l, &ref, &names, mapped_count, status);

	/* store the response in the SMB stream */
	return lsa_io_r_lookup_sids("", &r_l, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_open_policy2)
{
	LSA_Q_OPEN_POL2 q_o;
	LSA_R_OPEN_POL2 r_o;
	LSA_SEC_QOS qos;

	ZERO_STRUCT(q_o);
	ZERO_STRUCT(r_o);

	q_o.attr.sec_qos = &qos;

	if (!lsa_io_q_open_pol2("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_open_policy2(&q_o.uni_server_name, &r_o.pol,
				       &q_o.attr, q_o.des_access);
	return lsa_io_r_open_pol2("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_open_policy)
{
	LSA_Q_OPEN_POL q_o;
	LSA_R_OPEN_POL r_o;

	ZERO_STRUCT(r_o);
	ZERO_STRUCT(q_o);

	if (!lsa_io_q_open_pol("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_open_policy(NULL, &r_o.pol,
				      &q_o.attr, q_o.des_access);
	return lsa_io_r_open_pol("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_enum_trust_dom)
{
	uint32 status;
	uint32 enum_context;
	uint32 num_doms = 0;
	UNISTR2 *uni_names = NULL;
	DOM_SID **sids = NULL;
	LSA_R_ENUM_TRUST_DOM r_e;
	LSA_Q_ENUM_TRUST_DOM q_e;
	BOOL ret;

	ZERO_STRUCT(r_e);
	ZERO_STRUCT(q_e);

	/* grab the enum trust domain context etc. */
	if (!lsa_io_q_enum_trust_dom("", &q_e, data, 0))
	{
		return False;
	}

	enum_context = q_e.enum_context;

	status = _lsa_enum_trust_dom(&q_e.pol, &enum_context, &num_doms,
				     &uni_names, &sids);

	make_r_enum_trust_dom(&r_e, enum_context,
			      num_doms, uni_names, sids, status);

	/* store the response in the SMB stream */
	ret = lsa_io_r_enum_trust_dom("", &r_e, rdata, 0);

	/* free names and sids */
	free_sid_array(num_doms, sids);
	safe_free(uni_names);

	return ret;
}

RPC_SERVER_STUB_DECLARE(api_lsa_query_info)
{
	LSA_Q_QUERY_INFO q_i;
	LSA_R_QUERY_INFO r_i;

	ZERO_STRUCT(q_i);
	ZERO_STRUCT(r_i);

	/* grab the info class and policy handle */
	if (!lsa_io_q_query("", &q_i, data, 0))
	{
		return False;
	}

	r_i.status = _lsa_query_info_pol(&q_i.pol, q_i.info_class,
					 &r_i.dom);

	if (r_i.status == NT_STATUS_NOPROBLEMO)
	{
		/* set up the LSA QUERY INFO response */

		r_i.undoc_buffer = 1;
		r_i.info_class = q_i.info_class;
	}

	/* store the response in the SMB stream */
	return lsa_io_r_query("", &r_i, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_set_info)
{
	LSA_Q_SET_INFO q_i;
	ZERO_STRUCT(q_i);

	if (!lsa_io_q_set_info("", &q_i, data, 0))
	{
		return False;
	}

	return False;
}

RPC_SERVER_STUB_DECLARE(api_lsa_lookup_sids)
{
	LSA_Q_LOOKUP_SIDS q_l;
	ZERO_STRUCT(q_l);

	/* grab the info class and policy handle */
	if (!lsa_io_q_lookup_sids("", &q_l, data, 0))
	{
		return False;
	}

	/* construct reply.  return status is always 0x0 */
	return lsa_reply_lookup_sids(&q_l, rdata);
}

RPC_SERVER_STUB_DECLARE(api_lsa_lookup_names)
{
	LSA_Q_LOOKUP_NAMES q_l;
	LSA_R_LOOKUP_NAMES r_l;
	DOM_R_REF ref;
	DOM_RID2 *rids;
	uint32 mapped_count = 0;
	uint32 status;
	uint32 ret;

	ZERO_STRUCT(q_l);
	ZERO_STRUCT(r_l);
	ZERO_STRUCT(ref);
	rids = NULL;

	/* grab the info class and policy handle */
	if (!lsa_io_q_lookup_names("", &q_l, data, 0))
	{
		return False;
	}

	SMB_ASSERT_ARRAY(q_l.uni_name, q_l.num_entries);

	status = _lsa_lookup_names(&q_l.pol,
				   q_l.num_entries, q_l.uni_name,
				   &ref, &rids, &mapped_count);

	make_reply_lookup_names(&r_l, &ref, q_l.num_entries, rids,
				mapped_count, status);

	/* store the response in the SMB stream */
	ret = lsa_io_r_lookup_names("", &r_l, rdata, 0);

	safe_free(rids);

	return ret;
}

RPC_SERVER_STUB_DECLARE(api_lsa_close)
{
	LSA_R_CLOSE r_c;
	LSA_Q_CLOSE q_c;

	ZERO_STRUCT(q_c);
	ZERO_STRUCT(r_c);

	if (!lsa_io_q_close("", &q_c, data, 0))
	{
		return False;
	}

	r_c.pol = q_c.pol;	/* in/out */
	r_c.status = _lsa_close(&r_c.pol);
	return lsa_io_r_close("", &r_c, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_create_secret)
{
	LSA_R_CREATE_SECRET r_o;
	LSA_Q_CREATE_SECRET q_o;

	ZERO_STRUCT(q_o);
	ZERO_STRUCT(r_o);

	if (!lsa_io_q_create_secret("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_create_secret(&q_o.pol,
					&q_o.uni_secret, q_o.des_access,
					&r_o.pol);
	return lsa_io_r_create_secret("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_set_secret)
{
	LSA_Q_SET_SECRET q_o;
	LSA_R_SET_SECRET r_o;
	STRING2 *val = NULL;

	ZERO_STRUCT(r_o);
	ZERO_STRUCT(q_o);

	if (!lsa_io_q_set_secret("", &q_o, data, 0))
	{
		return False;
	}

	if (q_o.value.ptr_secret)
		val = &q_o.value.enc_secret;

	r_o.status = _lsa_set_secret(&q_o.pol, val, q_o.unknown);

	return lsa_io_r_set_secret("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_query_secret)
{
	LSA_R_QUERY_SECRET r_o;
	LSA_Q_QUERY_SECRET q_o;
	NTTIME *curtim = NULL;
	NTTIME *oldtim = NULL;
	STRING2 *curval = NULL;
	STRING2 *oldval = NULL;

	ZERO_STRUCT(r_o);
	ZERO_STRUCT(q_o);

	if (!lsa_io_q_query_secret("", &q_o, data, 0))
	{
		return False;
	}

	/* HACK! */
	if (q_o.sec.curinfo.ptr_value != 0)
		curval = &q_o.sec.curinfo.value.enc_secret;
	if (q_o.sec.curinfo.ptr_update != 0)
		curtim = &q_o.sec.curinfo.last_update;
	if (q_o.sec.oldinfo.ptr_value != 0)
		oldval = &q_o.sec.oldinfo.value.enc_secret;
	if (q_o.sec.oldinfo.ptr_update != 0)
		oldtim = &q_o.sec.oldinfo.last_update;

	r_o.status = _lsa_query_secret(&q_o.pol,
				       curval, curtim, oldval, oldtim);

	memcpy(&r_o.sec, &q_o.sec, sizeof(r_o.sec));	/* urgh! HACK! */
	if (r_o.sec.curinfo.ptr_value != 0)	/* MORE HACK! */
	{
		r_o.sec.curinfo.value.ptr_secret = 1;
		make_strhdr2(&r_o.sec.curinfo.value.hdr_secret,
			     r_o.sec.curinfo.value.enc_secret.str_str_len,
			     r_o.sec.curinfo.value.enc_secret.str_max_len, 1);
	}
	if (r_o.sec.oldinfo.ptr_value != 0)	/* MORE HACK! */
	{
		r_o.sec.curinfo.value.ptr_secret = 1;
		make_strhdr2(&r_o.sec.oldinfo.value.hdr_secret,
			     r_o.sec.oldinfo.value.enc_secret.str_str_len,
			     r_o.sec.oldinfo.value.enc_secret.str_max_len, 1);
	}

	return lsa_io_r_query_secret("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_open_secret)
{
	LSA_R_OPEN_SECRET r_o;
	LSA_Q_OPEN_SECRET q_o;

	ZERO_STRUCT(r_o);
	ZERO_STRUCT(q_o);

	if (!lsa_io_q_open_secret("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_open_secret(&q_o.pol,
				      &q_o.uni_secret, q_o.des_access,
				      &r_o.pol);
	return lsa_io_r_open_secret("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_enum_privs)
{
	LSA_Q_ENUM_PRIVS q_o;
	LSA_R_ENUM_PRIVS r_o;
	BOOL ret;

	ZERO_STRUCT(r_o);
	ZERO_STRUCT(q_o);

	if (!lsa_io_q_enum_privs("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_enum_privs(&q_o.pol, q_o.unk0, q_o.unk1,
				     &r_o.count, &r_o.privs);

	r_o.count2 = r_o.count1 = r_o.count;
	r_o.ptr = (r_o.count != 0 ? 1 : 0);

	ret = lsa_io_r_enum_privs("", &r_o, rdata, 0);

	safe_free(r_o.privs);

	return ret;
}

RPC_SERVER_STUB_DECLARE(api_lsa_priv_get_dispname)
{
	LSA_Q_PRIV_GET_DISPNAME q_i;
	LSA_R_PRIV_GET_DISPNAME r_i;
	UNISTR2 *uni_disp = NULL;

	ZERO_STRUCT(q_i);
	ZERO_STRUCT(r_i);

	if (!lsa_io_q_priv_get_dispname("", &q_i, data, 0))
	{
		return False;
	}

	r_i.status = _lsa_priv_get_dispname(&q_i.pol,
					&q_i.name, q_i.lang_id, q_i.lang_id_sys,
					&uni_disp, &r_i.lang_id);

	if (uni_disp)
	{
		r_i.ptr_info = 1;
		copy_unistr2(&r_i.desc, uni_disp);
		unistr2_free(uni_disp);
	}

	return lsa_io_r_priv_get_dispname("", &r_i, rdata, 0);
}


RPC_SERVER_STUB_DECLARE(api_lsa_create_trust_dom)
{
	LSA_Q_CREATE_TRUST_DOM q_o;
	MISC_R_HANDLE_STATUS r_o;
	BOOL ret;

	ZERO_STRUCT(q_o);

	if (! lsa_io_q_create_trust_dom("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_create_trust_dom(&q_o.hnd,
					   &q_o.trust_info,
					   q_o.des_access,
					   &r_o.hnd);

	ret = misc_io_r_handle_status("", &r_o, rdata, 0);

	return ret;
}


RPC_SERVER_STUB_DECLARE(api_lsa_open_trusted_dom)
{
	LSA_Q_OPEN_TRUSTED_DOM q_o;
	LSA_R_OPEN_TRUSTED_DOM r_o;

	ZERO_STRUCT(r_o);
	ZERO_STRUCT(q_o);

	if (!lsa_io_q_open_trusted_dom("", &q_o, data, 0))
	{
		return False;
	}

	r_o.status = _lsa_open_trusted_dom(&q_o.hnd,
					   &q_o.sid.sid, q_o.des_access,
					   &r_o.hnd);

	return lsa_io_r_open_trusted_dom("", &r_o, rdata, 0);
}

RPC_SERVER_STUB_DECLARE(api_lsa_delete_object)
{
	LSA_R_CLOSE r_c;
	LSA_Q_CLOSE q_c;

	ZERO_STRUCT(q_c);
	ZERO_STRUCT(r_c);

	if (!lsa_io_q_close("lsa_delete_object", &q_c, data, 0))
	{
		return False;
	}

	r_c.pol = q_c.pol;	/* in/out */

	r_c.status = _lsa_delete_object(&r_c.pol);

	return lsa_io_r_close("lsa_delete_object", &r_c, rdata, 0);
}

/***************************************************************************
 \PIPE\ntlsa commands
 ***************************************************************************/
static const struct api_struct api_lsa_cmds[] = {
	{"LSA_OPENPOLICY2", LSA_OPENPOLICY2, api_lsa_open_policy2},
	{"LSA_OPENPOLICY", LSA_OPENPOLICY, api_lsa_open_policy},
	{"LSA_QUERYINFOPOLICY", LSA_QUERYINFOPOLICY, api_lsa_query_info},
	{"LSA_QUERY_INFO2", LSA_QUERY_INFO2, NULL},
	{"LSA_ENUMTRUSTDOM", LSA_ENUMTRUSTDOM, api_lsa_enum_trust_dom},
	{"LSA_CLOSE", LSA_CLOSE, api_lsa_close},
	{"LSA_OPENSECRET", LSA_OPENSECRET, api_lsa_open_secret},
	{"LSA_CREATESECRET", LSA_CREATESECRET, api_lsa_create_secret},
	{"LSA_QUERYSECRET", LSA_QUERYSECRET, api_lsa_query_secret},
	{"LSA_SETSECRET", LSA_SETSECRET, api_lsa_set_secret},
	{"LSA_LOOKUPSIDS", LSA_LOOKUPSIDS, api_lsa_lookup_sids},
	{"LSA_LOOKUPNAMES", LSA_LOOKUPNAMES, api_lsa_lookup_names},
	{"LSA_SET_INFO", LSA_SET_INFO, api_lsa_set_info},
	{"LSA_ENUM_PRIVS", LSA_ENUM_PRIVS, api_lsa_enum_privs},
	{"LSA_PRIV_GET_DISPNAME", LSA_PRIV_GET_DISPNAME, api_lsa_priv_get_dispname},
	{"LSA_PRIV_GET_VALUE", LSA_PRIV_GET_VALUE, NULL},
	{"LSA_PRIV_GET_SIDS", LSA_PRIV_GET_SIDS, NULL},
	{"LSA_SID_GET_PRIVS", LSA_SID_GET_PRIVS, NULL},
	{"LSA_SID_ADD_PRIVS", LSA_SID_ADD_PRIVS, NULL},
	{"LSA_SID_REM_PRIVS", LSA_SID_REM_PRIVS, NULL},
	{"LSA_QUERYSECOBJECT", LSA_QUERYSECOBJECT, NULL},
	{"LSA_CREATE_TRUST_DOM", LSA_CREATE_TRUST_DOM, api_lsa_create_trust_dom},
	{"LSA_OPEN_TRUSTED_DOM", LSA_OPEN_TRUSTED_DOM, api_lsa_open_trusted_dom},
	{"LSA_GET_TRUST_INFO", LSA_GET_TRUST_INFO, NULL},
	{"LSA_GET_TRUST_INFO2", LSA_GET_TRUST_INFO2, NULL},
	{"LSA_DELETE", LSA_DELETE, NULL},
	{"LSA_DELETE_OBJECT", LSA_DELETE_OBJECT, api_lsa_delete_object},
	{NULL, 0, NULL}
};

/***************************************************************************
 api_ntLsarpcTNP
 ***************************************************************************/
BOOL api_ntlsa_rpc(rpcsrv_struct * p)
{
	return api_rpcTNP(p, "api_ntlsa_rpc", api_lsa_cmds);
}
