/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Password and authentication handling
   Copyright (C) Jeremy Allison 1996-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-1998
      
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "sids.h"
#include "samdb.h"
#include "others.h"


/*
 * NOTE. All these functions are abstracted into a structure
 * that points to the correct function for the selected database. JRA.
 *
 * NOTE.  for the get/mod/add functions, there are two sets of functions.
 * one supports struct sam_passwd, the other supports struct smb_passwd.
 * for speed optimisation it is best to support both these sets.
 * 
 * it is, however, optional to support one set but not the other: there
 * is conversion-capability built in to passdb.c, and run-time error
 * detection for when neither are supported.
 * 
 * password database writers are recommended to implement the sam_passwd
 * functions in a first pass, as struct sam_passwd contains more
 * information, needed by the NT Domain support.
 * 
 * an API writer is expected to create either one set (struct smb_passwd) or
 * the other (struct sam_passwd) OR both, and optionally also to write display
 * info routines * (struct sam_disp_info).  functions which the API writer
 * chooses NOT to write must be wrapped in conversion functions (pwdb_x_to_y)
 * such that API users can call any function and still get valid results.
 *
 * the password API does NOT fill in the gaps if you set an API function
 * to NULL: it will deliberately attempt to call the NULL function.
 *
 */

static struct smb_passdb_ops *pwdb_ops;

/***************************************************************
 Initialise the password db operations.
***************************************************************/

BOOL initialise_password_db(void)
{
#define TRY(fn) if (pwdb_ops) return True; pwdb_ops = fn()
	/* order is significant - we can link in more than one db */
	TRY(nisplus_initialise_password_db);
	TRY(nt5ldap_initialise_password_db);
	TRY(ldap_initialise_password_db);
	TRY(mysql_initialise_password_db);
	TRY(file_initialise_password_db);
#undef TRY

	return pwdb_ops != NULL;
}



/************************************************************************
 Routine to del an entry from the smb passwd file.
*************************************************************************/

BOOL del_smbpwd_entry(uint32 rid)
{
	if (pwdb_ops->del_smbpwd_entry)
	{
		return pwdb_ops->del_smbpwd_entry(rid);
	}
	DEBUG(1, ("WARNING: del_smbpwd_entry not implemented in "
		  "current backend\n"));
	return False;
}

/*************************************************************
 initialises a struct smb_passwd.
 **************************************************************/
void pwdb_init_smb(struct smb_passwd *user)
{
	if (user == NULL) return;
	ZERO_STRUCTP(user);
	user->pass_last_set_time    = (time_t)-1;
	user->unix_uid = (uid_t)-1;
	user->user_rid = 0xffffffff;
}

static void show_filled_entry(const struct smb_passwd *smb)
{
	/*
	 * The calling code has to use the right
	 * debug level and check it.
	 */
	if (smb == NULL)
	{
		DEBUGADD(0, ("smb entry: NULL\n"));
		return;
	}

	DEBUGADD(0, ("smb entry: unix `%s' nt `%s' unix %d nt 0x%x\n",
	             smb->unix_name != NULL ? smb->unix_name : "NULL",
	             smb->nt_name   != NULL ? smb->nt_name   : "NULL",
	             smb->unix_uid, smb->user_rid));
}


/*************************************************************
 fills in missing details.  one set of details _must_ exist.
 **************************************************************/
struct smb_passwd *pwdb_smb_map_names(struct smb_passwd *smb)
{
	DOM_NAME_MAP gmep;
	BOOL found = False;
	DOM_SID sid;
	static fstring unix_name;
	static fstring nt_name;

	if (smb == NULL)
	{
		DEBUG(10,("pwdb_smb_map_names: NULL\n"));
		return NULL;
	}

	if (DEBUGLVL(10))
	{
		DEBUG(10, ("pwdb_smb_map_names: At entry:\n"));
		show_filled_entry(smb);
	}

	if (smb->unix_name == NULL && smb->nt_name == NULL &&
	    smb->unix_uid == (uid_t)-1 && smb->user_rid == 0xffffffff)
	{
		DEBUG(2, ("pwdb_smb_map_names: No useful info provided\n"));
		return NULL;
	}

	if (smb->unix_name != NULL && smb->nt_name != NULL &&
	    smb->unix_uid != (uid_t)-1 && smb->user_rid != 0xffffffff)
	{
		return smb;
	}

	if (!found && smb->unix_name != NULL)
	{
		found = lookupsmbpwnam(smb->unix_name, &gmep);
	}
	if (!found && smb->unix_uid  != (uid_t)-1)
	{
		found = lookupsmbpwuid(smb->unix_uid , &gmep);
	}

	if (!found)
	{
		sid_copy(&sid, &global_sam_sid);
		sid_append_rid(&sid, smb->user_rid);
	}

	if (!found && smb->user_rid != 0xffffffff)
	{
		found = lookupsmbpwsid  (&sid        , &gmep);
	}
	if (!found && smb->nt_name  != NULL)
	{
		found = lookupsmbpwntnam(smb->nt_name, &gmep);
	}

	if (!found)
	{
		if(DEBUGLVL(1))
		{
			DEBUGADD(1, ("WARNING: your passdb entry could not "
				     "be filled with correct info, please "
				     "check it:\n"));
			show_filled_entry(smb);
		}
		return NULL;
	}

	if (!sid_front_equal(&global_sam_sid, &gmep.sid))
	{
		fstring sid_str;
		sid_to_string(sid_str, &gmep.sid);
		DEBUG(0,("UNIX User %s Primary Group is in the wrong domain! %s\n",
		          smb->unix_name, sid_str));
		return NULL;
	}

	fstrcpy(unix_name, gmep.unix_name);
	fstrcpy(nt_name  , gmep.nt_name  );
	if (smb->unix_name == NULL      ) smb->unix_name = unix_name;
	if (smb->nt_name   == NULL      ) smb->nt_name   = nt_name  ;
	if (smb->unix_uid  == (uid_t)-1 ) smb->unix_uid  = (uid_t)gmep.unix_id;
	if (smb->user_rid  == 0xffffffff) sid_split_rid(&gmep.sid, &smb->user_rid);

	if (DEBUGLVL(10))
	{
		DEBUG(10, ("pwdb_smb_map_names: At exit:\n"));
		show_filled_entry(smb);
	}

	return smb;
}
