/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   NT Domain Authentication SMB / MSRPC client
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "rpcclient.h"
#include "nterr.h"

extern FILE* out_hnd;

/*
 * keys.  of the form:
 * ----
 *
 * [HKLM]|[HKU]|[HKCR]\[parent_keyname]\[subkey]|[value]
 *
 * reg_get_subkey() splits this down into:
 * [HKLM]|[HKU]|[HKCR]\[parent_keyname_components] and [subkey]|[value]
 *
 * reg_connect() splits the left side down further into:
 * [HKLM]|[HKU]|[HKCR] and [parent_keyname_components].
 *
 * HKLM is short for HKEY_LOCAL_MACHINE
 * HKCR is short for HKEY_CLASSES_ROOT
 * HKU  is short for HKEY_USERS
 * HKP  is short for HKEY_PERFORMANCE_DATA
 *
 * oh, and HKEY stands for "Hive Key".
 *
 */

/*****************************************************************
 splits out the last subkey of a key
 *****************************************************************/
static void reg_get_subkey(char *key_name, char *subkey_name)
{
	/* Interestingly, "\" is a legal character in reg value names,
	 * but (for obvious reasons) not in reg key names.
	 *
	 * To deal with this, we split on either the first "\\" or the
	 * last "\".  If you type "\\" for the separator, the value can
	 * have additional "\" characters.
	 */
	char *s = strstr(key_name, "\\\\");
	if (s)
	{
		*s++ = '\0';
	}
	else
	{
		s = strrchr(key_name, '\\');
	}
	if (s)
	{
		*s++ = '\0';
		fstrcpy(subkey_name, s);
	}
	else
	{
		subkey_name[0] = '\0';
	}
}


static NTSTATUS reg_display_key(POLICY_HND *unused_hnd, int val, const char *full_keyname, int num)
{
	switch (val)
	{
		case 0:
		{
			/* initialisation */
			report(out_hnd, "Key Name:\t%s\n", full_keyname);
			break;
		}
		case 1:
		{
			/* subkeys initialisation */
			if (num > 0)
			{
				report(out_hnd,"Subkeys\n");
				report(out_hnd,"-------\n");
			}
			break;
		}
		case 2:
		{
			/* values initialisation */
			if (num > 0)
			{
				report(out_hnd,"Key Values\n");
				report(out_hnd,"----------\n");
			}
			break;
		}
		case 3:
		default:
		{
			/* clean-up */
			break;
		}
	}
	return NT_STATUS_OK;
}

void split_server_keyname(char *srv_name, char *key, const char* arg)
{
	pstrcpy(key, arg);

	if (strnequal("\\\\", key, 2))
	{
		char *p = strchr(&key[2], '\\');
		if (p == NULL)
		{
			key[0] = 0;
			return;
		}

		*p = 0;

		fstrcpy(srv_name, key);
		pstrcpy(key, &arg[strlen(srv_name)+1]);
	}
}

/****************************************************************************
nt registry enum
****************************************************************************/
static NTSTATUS msrpc_reg_do_subkeys(POLICY_HND *key_pol,
				     const char *full_keyname,
				     uint32 num_subkeys,
				     RegKeyFn reg_fn,
				     RegSubKeyFn reg_key_fn)
{
	NTSTATUS status = NT_STATUS_OK;
	uint32 i;

	if (! reg_key_fn)
		return NT_STATUS_OK;

	if (reg_fn)
	{
		status = reg_fn(key_pol, 1, full_keyname, num_subkeys);
		if (! NT_STATUS_IS_OK(status))
			return status;
	}

	for (i = 0; i < num_subkeys; i++)
	{
		/*
		 * enumerate key
		 */

		fstring enum_name;
		uint32 enum_unk1;
		uint32 enum_unk2;
		time_t key_mod_time;

		reg_unknown_1a(key_pol, NULL);

		/* enum key */
		if (! reg_enum_key(key_pol, i, enum_name,
				   &enum_unk1, &enum_unk2,
				   &key_mod_time))
		{
			status = NT_STATUS_UNSUCCESSFUL;
			break;
		}

		status = reg_key_fn(key_pol, full_keyname, enum_name, key_mod_time);
		if (! NT_STATUS_IS_OK(status))
			break;
	}
	return status;
}


static NTSTATUS msrpc_reg_do_subvalues(POLICY_HND *key_pol,
				       const char *full_keyname,
				       uint32 num_values,
				       uint32 max_valnamelen,
				       uint32 max_valbufsize,
				       RegKeyFn reg_fn,
				       RegSubValueFn reg_val_fn)
{
	NTSTATUS status = NT_STATUS_OK;
	uint32 i;

	if (! reg_val_fn)
		return NT_STATUS_OK;

	if (reg_fn)
	{
		status = reg_fn(key_pol, 2, full_keyname, num_values);
		if (! NT_STATUS_IS_OK(status))
			return status;
	}

	for (i = 0; i < num_values; i++)
	{
		uint32 val_type;
		BUFFER2 value;
		fstring val_name;

		ZERO_STRUCT(value);

		reg_unknown_1a(key_pol, NULL);

		if (! reg_enum_val(key_pol,
				   i, max_valnamelen, max_valbufsize,
				   val_name, &val_type, &value))
		{
			status = NT_STATUS_UNSUCCESSFUL;
			break;
		}

		status = reg_val_fn(key_pol, full_keyname, val_name, val_type, &value);

		misc_buffer2_free(&value);

		if (! NT_STATUS_IS_OK(status))
			break;
	}

	return status;
}


NTSTATUS msrpc_reg_enum_key(POLICY_HND *key_pol,
			    const char *full_keyname,
			    BOOL subvalues_first,
			    RegKeyFn reg_fn,
			    RegSubKeyFn reg_key_fn,
			    RegSubValueFn reg_val_fn)
{
	NTSTATUS status;

	/*
	 * query key info
	 */

	fstring key_class;
	uint32 max_class_len = 0;
	uint32 num_subkeys = 0;
	uint32 max_subkeylen;
	uint32 max_subkeysize;
	uint32 num_values = 0;
	uint32 max_valnamelen;
	uint32 max_valbufsize;
	uint32 sec_desc;
	NTTIME mod_time;

	if (! reg_query_key(key_pol,
			    key_class, &max_class_len,
			    &num_subkeys, &max_subkeylen, &max_subkeysize,
			    &num_values, &max_valnamelen, &max_valbufsize,
			    &sec_desc, &mod_time))
	{
		return NT_STATUS_UNSUCCESSFUL;
	}

	if (reg_fn)
	{
		reg_fn(key_pol, 0, full_keyname, 0);
	}

	if (! subvalues_first)
	{
		status = msrpc_reg_do_subkeys(key_pol,
					      full_keyname,
					      num_subkeys,
					      reg_fn,
					      reg_key_fn);
		if (! NT_STATUS_IS_OK(status))
			return status;
	}

	status = msrpc_reg_do_subvalues(key_pol,
					full_keyname,
					num_values,
					max_valnamelen,
					max_valbufsize,
					reg_fn,
					reg_val_fn);
	if (! NT_STATUS_IS_OK(status))
		return status;

	if (subvalues_first)
	{
		status = msrpc_reg_do_subkeys(key_pol,
					      full_keyname,
					      num_subkeys,
					      reg_fn,
					      reg_key_fn);
		if (! NT_STATUS_IS_OK(status))
			return status;
	}

	if (reg_fn)
	{
		/*
		 * Or should this one be called anyway,
		 * no matter, if the previous stuff failed?
		 */
		reg_fn(key_pol, 3, full_keyname, 0);
	}

	return status;
}


NTSTATUS msrpc_reg_enum_keyname(const char *srv_name, const char *full_keyname,
				BOOL subvalues_first,
				RegKeyFn reg_fn,
				RegSubKeyFn reg_key_fn,
				RegSubValueFn reg_val_fn)
{
	NTSTATUS status;

	POLICY_HND key_pol;
	POLICY_HND pol_con;
	fstring key_name;

	DEBUG(5, ("msrpc_reg_enum_keyname: %s\n", full_keyname));

	/* open registry receive a policy handle */
	if (! reg_connect(srv_name, full_keyname, key_name,
			  SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
	{
		return NT_STATUS_UNSUCCESSFUL;
	}

	if (*key_name)
	{
		if (! reg_open_entry(&pol_con, key_name,
				     SEC_RIGHTS_MAXIMUM_ALLOWED, &key_pol))
		{
			status = NT_STATUS_UNSUCCESSFUL;
			goto out;
		}
	}
	else
	{
		key_pol = pol_con;
	}

	status = msrpc_reg_enum_key(&key_pol,
				    full_keyname,
				    subvalues_first,
				    reg_fn,
				    reg_key_fn,
				    reg_val_fn);

	/* close the handles */
	if (*key_name)
	{
		reg_close(&key_pol);
	}
    out:
	reg_close(&pol_con);

	return status;
}

static NTSTATUS reg_display_key_info(POLICY_HND *unused_hnd, const char *full_name,
				const char *name, time_t key_mod_time)
{
	display_reg_key_info(out_hnd, ACTION_HEADER   , name, key_mod_time);
	display_reg_key_info(out_hnd, ACTION_ENUMERATE, name, key_mod_time);
	display_reg_key_info(out_hnd, ACTION_FOOTER   , name, key_mod_time);
	return NT_STATUS_OK;
}

static NTSTATUS reg_display_val_info(POLICY_HND *unused_hnd, const char *full_name,
				const char* name,
				uint32 type,
				const BUFFER2 *const value)
{
	display_reg_value_info(out_hnd, ACTION_HEADER   , name, type, value);
	display_reg_value_info(out_hnd, ACTION_ENUMERATE, name, type, value);
	display_reg_value_info(out_hnd, ACTION_FOOTER   , name, type, value);
	return NT_STATUS_OK;
}

/****************************************************************************
nt registry enum
****************************************************************************/
NTSTATUS cmd_reg_enum(struct client_info *info, int argc, char *argv[])
{
	pstring full_keyname;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regenum <key_name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, full_keyname, argv[1]);

	return msrpc_reg_enum_keyname(srv_name, full_keyname, False,
				      reg_display_key,
				      reg_display_key_info,
				      reg_display_val_info);
}

/****************************************************************************
nt registry query value info
****************************************************************************/
NTSTATUS cmd_reg_query_info(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND key_pol;
	POLICY_HND pol_con;
	pstring full_keyname;
	pstring keyname;
	fstring key_name;
	fstring val_name;

	/*
	 * query value info
	 */

	BUFFER2 buf;
	uint32 type;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regvalinfo value_name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, full_keyname, argv[1]);

	pstrcpy(keyname, full_keyname);
	reg_get_subkey(keyname, val_name);

	if (keyname[0] == 0 || val_name[0] == 0)
	{
		report(out_hnd, "invalid value name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, keyname, key_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*key_name)
	{
		/* open an entry */
		if (!reg_open_entry(&pol_con, key_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &key_pol))
			goto out1;
	}
	else
	{
		key_pol = pol_con;
	}

	buf.buf_max_len = 102400; /* requested maximum size */

	/* query it */
	if (reg_query_info(&key_pol, val_name, &type, &buf))
	{
		reg_display_val_info(&key_pol, full_keyname, val_name, type, &buf);
		misc_buffer2_free(&buf);
		ret = NT_STATUS_OK;
	}

	if (*key_name)
	{
		reg_close(&key_pol);
	}
out1:
	reg_close(&pol_con);
out:
	DEBUG(5,("cmd_reg_query: %s\n",
		 NT_STATUS_IS_OK(ret) ? "query succeeded" : "query failed"));
	return ret;
}

/****************************************************************************
nt registry query key
****************************************************************************/
NTSTATUS cmd_reg_query_key(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND key_pol;
	POLICY_HND pol_con;
	pstring full_keyname;
	fstring key_name;

	/*
	 * query key info
	 */

	fstring key_class;
	uint32 key_class_len = 0;
	uint32 num_subkeys;
	uint32 max_subkeylen;
	uint32 max_subkeysize;
	uint32 num_values;
	uint32 max_valnamelen;
	uint32 max_valbufsize;
	uint32 sec_desc;
	NTTIME mod_time;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regquery key_name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, full_keyname, argv[1]);

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, full_keyname, key_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*key_name)
	{
		if (!reg_open_entry(&pol_con, key_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &key_pol))
			goto out1;
	}
	else
	{
		key_pol = pol_con;
	}

	if (!reg_query_key(&key_pol, key_class, &key_class_len,
			   &num_subkeys, &max_subkeylen, &max_subkeysize,
			   &num_values, &max_valnamelen, &max_valbufsize,
			   &sec_desc, &mod_time))
		goto out2;

	if (key_class_len != 0)
	{
		if (!reg_query_key(&key_pol, key_class, &key_class_len,
				   &num_subkeys, &max_subkeylen, &max_subkeysize,
				   &num_values, &max_valnamelen, &max_valbufsize,
				   &sec_desc, &mod_time))
			goto out2;
	}

	report(out_hnd,"Registry Query Info Key\n");
	report(out_hnd,"key class: %s\n", key_class);
	report(out_hnd,"subkeys, max_len, max_size: %d %d %d\n", num_subkeys, max_subkeylen, max_subkeysize);
	report(out_hnd,"vals, max_len, max_size: 0x%x 0x%x 0x%x\n", num_values, max_valnamelen, max_valbufsize);
	report(out_hnd,"sec desc: 0x%x\n", sec_desc);
	report(out_hnd,"mod time: %s\n", http_timestring(nt_time_to_unix(&mod_time)));
	ret = NT_STATUS_OK;

out2:
	if (*key_name)
		reg_close(&key_pol);
out1:
	reg_close(&pol_con);
out:
	DEBUG(5,("cmd_reg_query: %s\n",
		 NT_STATUS_IS_OK(ret) ? "query succeeded" : "query failed"));
	return ret;
}

extern struct field_info reg_val_types[];
/****************************************************************************
nt registry create value
****************************************************************************/
NTSTATUS cmd_reg_create_val(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND parent_pol;
	POLICY_HND pol_con;
	pstring keyname;
	fstring parent_name;
	fstring val_name;
	uint32 val_type;
	BUFFER3 value;

#if 0
	uint32 unk_0;
	uint32 unk_1;
	/* query it */
	reg_query_info(&val_pol, type, &unk_0, &unk_1);
#endif

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	argc--;
	argv++;

	if (argc < 2)
		goto usage;

	split_server_keyname(srv_name, keyname, argv[0]);
	argc--;
	argv++;

	reg_get_subkey(keyname, val_name);

	if (keyname[0] == 0)
	{
		report(out_hnd, "invalid key name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (sscanf(argv[0], "%i", &val_type) < 1)
		val_type = str_to_enum_field(argv[0], reg_val_types,
					     REGVAL_INVALID_TYPE);
	argc--;
	argv++;

	if (argc < 1 && val_type != REGVAL_NONE && val_type != REGVAL_SZ_MULTI)
	{
	usage:
		cmd_help_current(info);
		return NT_STATUS_INVALID_PARAMETER;
	}

	switch (val_type)
	{
		case REGVAL_NONE:
		{
			make_buffer3_str(&value, "", 0);
			break;
		}
		case REGVAL_SZ:
		case REGVAL_EXPAND_SZ:
		{
			make_buffer3_str(&value, argv[0], strlen(argv[0])+1);
			break;
		}
		case REGVAL_FULL_RES_DESCR:
		case REGVAL_RES_REQ_LIST:
		case REGVAL_BINARY:
		default:
		{
			make_buffer3_hex(&value, argc, argv);
			break;
		}
		case REGVAL_DWORD:
		{
			make_buffer3_uint32(&value, get_number(argv[0]));
			break;
		}
		case REGVAL_SZ_MULTI:
		{
			make_buffer3_multistr(&value, argc, argv);
			break;
		}
		case REGVAL_INVALID_TYPE:
		{
			goto usage;
		}
	}

	DEBUG(10,("key data:\n"));
	dump_data(10, (char *)value.buffer, value.buf_len);

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, keyname, parent_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*parent_name)
	{
		if (!reg_open_entry(&pol_con, parent_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &parent_pol))
			goto out1;
	}
	else
	{
		parent_pol = pol_con;
	}

	if (reg_create_val(&parent_pol, val_name, val_type, &value) &&
	    reg_flush_key(&parent_pol))
	{
		ret = NT_STATUS_OK;
	}

	if (*parent_name)
		reg_close(&parent_pol);
out1:
	reg_close(&pol_con);
out:
	if (NT_STATUS_IS_OK(ret))
	{
		DEBUG(5,("cmd_reg_create_val: query succeeded\n"));
		report(out_hnd,"OK\n");
	}
	else
	{
		DEBUG(5,("cmd_reg_create_val: query failed\n"));
	}
	return ret;
}

/****************************************************************************
nt registry delete value
****************************************************************************/
NTSTATUS cmd_reg_delete_val(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND parent_pol;
	POLICY_HND pol_con;
	pstring keyname;
	fstring parent_name;
	fstring val_name;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regdelete <val_name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, keyname, argv[1]);

	reg_get_subkey(keyname, val_name);

	if (keyname[0] == 0 || val_name[0] == 0)
	{
		report(out_hnd, "invalid key name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, keyname, parent_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*val_name)
	{
		if (!reg_open_entry(&pol_con, parent_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &parent_pol))
			goto out1;
	}
	else
	{
		parent_pol = pol_con;
	}

	if (reg_delete_val(&parent_pol, val_name) &&
	    reg_flush_key(&parent_pol))
	{
		ret = NT_STATUS_OK;
	}

	if (*val_name)
		reg_close(&parent_pol);
out1:
	reg_close(&pol_con);
out:
	if (NT_STATUS_IS_OK(ret))
	{
		DEBUG(5,("cmd_reg_delete_val: query succeeded\n"));
		report(out_hnd,"OK\n");
	}
	else
	{
		DEBUG(5,("cmd_reg_delete_val: query failed\n"));
	}
	return ret;
}

/****************************************************************************
nt registry delete key
****************************************************************************/
NTSTATUS cmd_reg_delete_key(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND parent_pol;
	POLICY_HND pol_con;
	pstring parent_name;
	fstring key_name;
	fstring subkey_name;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regdeletekey <key_name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, parent_name, argv[1]);

	reg_get_subkey(parent_name, subkey_name);

	if (parent_name[0] == 0 || subkey_name[0] == 0)
	{
		report(out_hnd, "invalid key name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, parent_name, key_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*key_name)
	{
		/* open an entry */
		if (!reg_open_entry(&pol_con, key_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &parent_pol))
			goto out1;
	}
	else
	{
		parent_pol = pol_con;
	}

	if (reg_delete_key(&parent_pol, subkey_name) &&
	    reg_flush_key(&parent_pol))
	{
		ret = NT_STATUS_OK;
	}

	if (*key_name)
		reg_close(&parent_pol);
out1:
	reg_close(&pol_con);
out:
	DEBUG(5,("cmd_reg_delete_key: %s\n",
		 NT_STATUS_IS_OK(ret) ? "succeeded" : "failed"));
	return ret;
}

/****************************************************************************
nt registry create key
****************************************************************************/
NTSTATUS cmd_reg_create_key(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND parent_pol;
	POLICY_HND key_pol;
	POLICY_HND pol_con;
	pstring parent_key;
	fstring parent_name;
	fstring key_name;
	fstring key_class;
	SEC_ACCESS sam_access;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regcreate <key_name> [key_class]\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, parent_key, argv[1]);

	reg_get_subkey(parent_key, key_name);

	if (parent_key[0] == 0 || key_name[0] == 0)
	{
		report(out_hnd, "invalid key name\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	if (argc > 2)
	{
		fstrcpy(key_class, argv[2]);
	}
	else
	{
		ZERO_ARRAY(key_class);
	}

	/* set access permissions */
	sam_access.mask = SEC_RIGHTS_READ;

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, parent_key, parent_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*parent_name)
	{
		if (!reg_open_entry(&pol_con, parent_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &parent_pol))
			goto out1;
	}
	else
	{
		parent_pol = pol_con;
	}

	/* create an entry */
	if (reg_create_key(&parent_pol, key_name, key_class,
			   &sam_access, &key_pol) &&
	    reg_flush_key(&parent_pol) &&
	    reg_close(&key_pol))
	{
		ret = NT_STATUS_OK;
	}

	if (*parent_name)
		reg_close(&parent_pol);
out1:
	reg_close(&pol_con);
out:
	if (NT_STATUS_IS_OK(ret))
	{
		DEBUG(5,("cmd_reg_create_key: query succeeded\n"));
		report(out_hnd,"OK\n");
	}
	else
	{
		DEBUG(5,("cmd_reg_create_key: query failed\n"));
	}
	return ret;
}

/****************************************************************************
nt registry security info
****************************************************************************/
NTSTATUS cmd_reg_test_key_sec(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND key_pol;
	POLICY_HND pol_con;
	pstring full_keyname;
	fstring key_name;

	/*
	 * security info
	 */

	uint32 sec_buf_size;
	SEC_DESC_BUF sec_buf;
	uint32 sec_info = 0x7;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "regtestkeysec <key_name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, full_keyname, argv[1]);

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, full_keyname, key_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*key_name)
	{
		if (!reg_open_entry(&pol_con, key_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &key_pol))
			goto out1;
	}
	else
	{
		key_pol = pol_con;
	}

	/* query key sec info.  first call sets sec_buf_size. */
	sec_buf_size = 1024;
	ZERO_STRUCT(sec_buf);

	if (!reg_get_key_sec(&key_pol, sec_info, &sec_buf_size, &sec_buf))
		goto out2;
	free_sec_desc_buf(&sec_buf);

	if (!reg_get_key_sec(&key_pol, sec_info, &sec_buf_size, &sec_buf))
		goto out2;

	if (sec_buf.len > 0 && sec_buf.sec != NULL)
	{
		print_sec_desc(sec_buf.sec, out_hnd, NULL, 0, NULL,
			       srv_name, NULL);

		if (reg_set_key_sec(&key_pol, sec_info, sec_buf_size, sec_buf.sec))
		{
			ret = NT_STATUS_OK;
		}
		free_sec_desc_buf(&sec_buf);
	}

out2:
	if (*key_name)
		reg_close(&key_pol);
out1:
	reg_close(&pol_con);
out:
	if (NT_STATUS_IS_OK(ret))
	{
		DEBUG(5,("cmd_reg_test2: query succeeded\n"));
		report(out_hnd,"Registry Test2\n");
	}
	else
	{
		DEBUG(5,("cmd_reg_test2: query failed\n"));
	}
	return ret;
}

/****************************************************************************
nt registry security info
****************************************************************************/
NTSTATUS cmd_reg_get_key_sec(struct client_info *info, int argc, char *argv[])
{
	NTSTATUS ret = NT_STATUS_UNSUCCESSFUL;

	POLICY_HND key_pol;
	POLICY_HND pol_con;
	pstring full_keyname;
	fstring key_name;

	/*
	 * security info
	 */

	uint32 sec_buf_size;
	SEC_DESC_BUF sec_buf;
	uint32 sec_info = 0x7;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (argc < 2)
	{
		report(out_hnd, "reggetsec <key_name>\n");
		return NT_STATUS_INVALID_PARAMETER;
	}

	split_server_keyname(srv_name, full_keyname, argv[1]);

	/* open registry receive a policy handle */
	if (!reg_connect(srv_name, full_keyname, key_name,
			 SEC_RIGHTS_MAXIMUM_ALLOWED, &pol_con))
		goto out;

	if (*key_name)
	{
		if (!reg_open_entry(&pol_con, key_name,
				    SEC_RIGHTS_MAXIMUM_ALLOWED, &key_pol))
			goto out1;
	}
	else
	{
		key_pol = pol_con;
	}

	/* query key sec info.  first call sets sec_buf_size. */
	sec_buf_size = 0;
	ZERO_STRUCT(sec_buf);

	if (!reg_get_key_sec(&key_pol, sec_info, &sec_buf_size, NULL))
		goto out2;
	free_sec_desc_buf(&sec_buf);

	if (!reg_get_key_sec(&key_pol, sec_info, &sec_buf_size, &sec_buf))
		goto out2;

	if (sec_buf.len > 0 && sec_buf.sec != NULL)
	{
		print_sec_desc(sec_buf.sec, out_hnd, NULL,
			       SECOBJ_REGKEY, NULL,
			       srv_name, NULL);

		free_sec_desc_buf(&sec_buf);
		ret = NT_STATUS_OK;
	}

out2:
	if (*key_name)
		reg_close(&key_pol);
out1:
	reg_close(&pol_con);
out:
	DEBUG(5,("cmd_reg_get_key_sec: %s\n",
		 NT_STATUS_IS_OK(ret) ? "succeeded" : "failed"));
	return ret;
}

/****************************************************************************
nt registry shutdown
****************************************************************************/
NTSTATUS cmd_reg_shutdown(struct client_info *info, int argc, char *argv[])
{
	fstring msg;
	uint32 timeout = 20;
	uint16 flgs = 0;
	int opt;

	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	msg[0] = 0;

	while ((opt = getopt(argc, argv,"fim:t:r-")) != EOF)
	{
		switch (opt)
		{
			case 'm':
			{
				fstrcpy(msg, optarg);
				break;
			}
			case 't':
			{
				timeout = atoi(optarg);
				break;
			}
			case 'r':
			{
				flgs |= 0x100;
				break;
			}
			case 'f':
			{
				flgs |= 0x001;
				break;
			}
			case '-':
			{
				if (strequal(optarg, "-reboot"))
				{
					flgs |= 0x100;
				}
				if (strequal(optarg, "-force-close"))
				{
					flgs |= 0x001;
				}
				break;
			}
		}
	}

	if (reg_shutdown(srv_name, msg, timeout, flgs))
	{
		DEBUG(5,("cmd_reg_shutdown: succeeded\n"));
		return NT_STATUS_OK;
	}
	DEBUG(5,("cmd_reg_shutdown: failed\n"));
	return NT_STATUS_UNSUCCESSFUL;
}

/****************************************************************************
abort a shutdown
****************************************************************************/
NTSTATUS cmd_reg_abort_shutdown(struct client_info *info, int argc, char *argv[])
{
	fstring srv_name;

	fstrcpy(srv_name, "\\\\");
	fstrcat(srv_name, info->dest_host);
	strupper(srv_name);

	if (reg_abort_shutdown(srv_name))
	{
		DEBUG(5,("cmd_reg_abort_shutdown: succeeded\n"));
		return NT_STATUS_OK;
	}
	DEBUG(5,("cmd_reg_abort_shutdown: failed\n"));
	return NT_STATUS_UNSUCCESSFUL;
}
