/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client
   Copyright (C) Andrew Tridgell              1994-2000
   Copyright (C) Luke Kenneth Casson Leighton 1996-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "nterr.h"
#include "rpcclient.h"
#include "tng_readline.h"

extern struct client_info cli_info;

/* Complete a remote registry enum */

#ifdef HAVE_READLINE
static uint32 reg_list_len = 0;
static char **reg_name = NULL;

static NTSTATUS reg_init(POLICY_HND *unused_hnd, int val, const char *full_keyname, int num)
{
	switch (val)
	{
		case 0:
		{
			free_char_array(reg_list_len, reg_name);
			reg_list_len = 0;
			reg_name = NULL;
			break;
		}
		default:
		{
			break;
		}
	}
	return NT_STATUS_OK;
}

static NTSTATUS reg_key_list(POLICY_HND *unused_hnd, const char *full_name,
			 const char *name, time_t key_mod_time)
{
	fstring key_name;
	slprintf(key_name, sizeof(key_name) - 1, "%s\\", name);
	add_chars_to_array(&reg_list_len, &reg_name, key_name);
	return NT_STATUS_OK;
}

static NTSTATUS reg_val_list(POLICY_HND *unused_hnd, const char *full_name,
			 const char *name, uint32 type, const BUFFER2 * value)
{
	fstring val_name;
	slprintf(val_name, sizeof(val_name) - 1, "%s ", name);
	add_chars_to_array(&reg_list_len, &reg_name, val_name);
	return NT_STATUS_OK;
}

extern char** cmd_argv;
extern uint32 cmd_argc;

static char *complete_regenum(const char *text, int state)
{
	pstring full_keyname;
	static uint32 i = 0;

	/* we manually add ' ' or '\\' to completions, so disable auto ' ' */
	rl_completion_append_character = '\0';
	if (state == 0)
	{
		fstring srv_name;
		fstrcpy(srv_name, "\\\\");
		fstrcat(srv_name, cli_info.dest_host);
		strupper(srv_name);

		if (cmd_argc >= 2 && cmd_argv != NULL && cmd_argv[1] != NULL)
		{
			char *sep;
			split_server_keyname(srv_name, full_keyname,
					     cmd_argv[1]);

			sep = strrchr(full_keyname, '\\');
			if (sep != NULL)
			{
				*sep = 0;
			}
		}

		/* Iterate all keys / values */
		if (! NT_STATUS_IS_OK(msrpc_reg_enum_keyname(srv_name,
							     full_keyname,
							     False,
							     reg_init,
							     reg_key_list,
							     reg_val_list)))
		{
			return NULL;
		}

		i = 0;
	}

	for (; i < reg_list_len; i++)
	{
		if (text == NULL || text[0] == 0 ||
		    strnequal(text, reg_name[i], strlen(text)))
		{
			char *name = strdup(reg_name[i]);
			i++;
			return name;
		}
	}

	return NULL;
}
#else
#define complete_regenum NULL
#endif /* READLINE */

/****************************************************************************
 This defines the commands supported by this client
 ****************************************************************************/
static const struct command_set reg_commands[] = 
{
	{
		"registry", NULL,
		"Commands to edit the registry",
	},
	{
		"registry key", NULL,
		"Commands to read / write registry keys",
	},
	{
		"registry value", NULL,
		"Commands to read / write registry values",
	},

	{
		"registry key enum",
		cmd_reg_enum,
		"<keyname> Enumerate subkeys and values of a key",
		{complete_regenum, NULL}
	},
	{
		"registry key query",
		cmd_reg_query_key,
		"<keyname> List the attributes of a key",
		{complete_regenum, NULL}
	},
	{
		"registry key create",
		cmd_reg_create_key,
		"<keyname> [<class>] Create a registry key",
		{complete_regenum, NULL}
	},
	{
		"registry key delete",
		cmd_reg_delete_key,
		"<keyname> Delete a registry key",
		{complete_regenum, NULL}
	},
	{
		"registry value query",
		cmd_reg_query_info,
		"<valname> List the data associated with a value",
		{complete_regenum, NULL}
	},
	{
		"registry value create",
		cmd_reg_create_val,
		"<name> <type> <data> Create a registry value\n"
		"  where <type> is one of:\n"
		"    REG_NONE      no data\n"
		"    REG_SZ        a string, which will be converted to unicode\n"
		"    REG_EXPAND_SZ like REG_SZ, but apps will interpolate %VARIABLE% strings\n"
		"    REG_BINARY    arbitrary binary data; input as hexadecimal, with leading 0s\n"
		"    REG_DWORD     a 32-bit integer\n"
		"    REG_SZ_MULTI  a list of strings, converted to unicode\n"
		"    REG_FULL_RESOURCE_DESCRIPTOR    treated like REG_BINARY\n"
		"    REG_RESOURCE_REQUIREMENTS_LIST  treated like REG_BINARY",
		{complete_regenum, NULL}
	},
	{
		"registry value delete",
		cmd_reg_delete_val,
		"<valname> Delete a registry value",
		{complete_regenum, complete_regenum}
	},
	{
		"registry key getsec",
		cmd_reg_get_key_sec,
		"<keyname> Query the security descriptor for a registry key",
		{complete_regenum, NULL}
	},
	{
		"registry key testsec",
		cmd_reg_test_key_sec,
		"<keyname> Test registry key security",
		{complete_regenum, NULL}
	},
	{
		"registry dump",
		cmd_reg_dump,
		"[args] <keyname> Dump registry key and its children\n"
/*		"  -t <type>   specify format of output: {nt4|nt5|samba} (default: nt4)" */
		"  -o <file>   output to <file>\n",
		{complete_regenum, NULL}
	},
/*
	{
		"registry restore",
		cmd_reg_restore,
		"[-t type] [file] Upload registry dump file\n"
		"  -t <type>   specify format of input: {nt4|nt5|samba} (default: auto)",
		{NULL, NULL}
	},
*/
	/* backward-compatible aliases for the above */
	{ "regenum", cmd_reg_enum, },
	{ "regdeletekey", cmd_reg_delete_key, },
	{ "regcreatekey", cmd_reg_create_key, },
	{ "regqueryval", cmd_reg_query_info, },
	{ "regquerykey", cmd_reg_query_key, },
	{ "regdeleteval", cmd_reg_delete_val, },
	{ "regcreateval", cmd_reg_create_val, },
	{ "reggetsec", cmd_reg_get_key_sec, },
	{ "regtestsec", cmd_reg_test_key_sec, },

	{
		"shutdown",
		cmd_reg_shutdown,
		"Remote Shutdown\n"
		"\t[-m message] [-t timeout] [-r or --reboot] [-f or --force-close]",
		{NULL, NULL}
	},
	{
		"abortshutdown",
		cmd_reg_abort_shutdown,
		"Abort Shutdown",
		{NULL, NULL}
	},

	{
		"",
	}
};

void add_reg_commands(void)
{
	add_command_set(reg_commands);
}
