/*
 * mkproto.c - dig out function prototypes from C files
 * Copyright 2001, 2003 by Peter Samuelson <peter@samba-tng.org>
 *
 * The design is based on the awk program 'mkproto.awk' and its shell
 * wrapper 'mkproto.sh', both by the Samba Team.  It has since diverged
 * somewhat; see the comment below for the current workings.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 * USA.
 */
/* MKPROTO: NONE			(do not remove this line) */

/*
 * Theory of Operation:
 *
 * Every .c source file is scanned for prototypes - basically, any
 * alphabetic character at the beginning of a line signals a possible
 * function declaration, and is searched for known return types.  Some
 * very rudimentary parsing is done to distinguish valid declarations
 * from other things that happen in column 1, and to extract multi-line
 * declarations.
 *
 * Any prototypes found are cached in a file under '.proto/', then
 * collected into various header files in the source tree.
 *
 * For any file whose prototypes should be put in 'include/proto.h', DO
 * NOTHING; this is the default.  To put a file's prototypes in another
 * header file, declare the header using a line beginning with the
 * following magic comment (but without the initial spaces): */
   /* MKPROTO: header_file_name [the rest of the line is ignored] */
/* If the header name includes at least one "/" character, it
 * is placed as-is in the source directory.  If not, it is placed under
 * "include/" in the source directory.
 *
 * If the specified header is "NONE" (see the line in this file),
 * prototypes are not extracted at all.
 *
 * To defined a new return type to be scanned, add the type to the
 * 'start_tokens' array.  To extract prototypes from a macro, add the
 * appropriate specification to the 'macro_tokens' array.
 */

/* the ifdefs are based on what is in include/includes.h */
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#ifndef HAVE_CONFIG_H	/* assume sane environment */
#  define HAVE_UNISTD_H
#  define HAVE_STRING_H
#  define HAVE_CTYPE_H
#  define HAVE_STRCHR
#endif

#ifdef HAVE_SYS_UNISTD_H
#include <sys/unistd.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif

#define BUFLEN 4096
#define LINELEN 512	/* would you believe some lines are >256 char? */
#define HEADER_TEMP ".tmp.header.%d"
#define PROTO_TEMP ".tmp.proto"
#define OUTPUT_FILE_MAGIC "/* MKPROTO:"
#define OUTPUT_FILE_NONE "NONE"

#define STRUCT_SIZE(s) (sizeof(s) / sizeof(*(s)))
#define ZERO_MTIME(st) (st.st_mtime = 0)
#define NEWER(s1,s2) (s1.st_mtime > s2.st_mtime)

static char *progname;
static char *default_header_name = "include/proto.h";
static char *source_dir = ".";
static char *proto_dir = ".proto";
static int force;
static char errbuf[LINELEN];

static char *start_tokens[] = {
  /* common culprits, in order of popularity */
  "BOOL", "void", "uint32", "int", "char", "struct",
  "enum winbindd_result", "smb_ucs2_t", "ssize_t", "const",
  "size_t", "files_struct", "time_t", "LOCAL_GRP",

  /* uncommon culprits, shortest-to-longest */
  "DIR", "FILE", "long", "pid_t", "off_t", "int32", "ino_t",
  "mode_t", "uint16", "DOM_SID", "UNISTR2", "unsigned", "TDB_DATA",
  "SMB_OFF_T", "DOMAIN_GRP", "JOB_INFO_2", "JOB_INFO_1",
  "TALLOC_CTX", "TDB_CONTEXT", "user_struct", "SMB_BIG_UINT",
  "pipes_struct", "hash_element", "GENERIC_LIST", "NT_DEVICEMODE",
  "PRINTER_INFO_2", "PRINTER_INFO_1", "cli_connect_fns",
  "connection_struct", "msrpc_service_fns", "SMB_STRUCT_DIRENT",
  "enum protocol_types", "enum remote_arch_types",
  "NT_USER_TOKEN", "smb_iconv_t", "SEC_ACL",
  "NTSTATUS", "WERROR",
};

static char *ignore_start_tokens[] = {
  "static", "extern",
};

static struct { char *token, *fmt; } macro_tokens[] = {
  /* once again in order of popularity */
  { "FN_GLOBAL_STRING",  "char *%s(void)" },
  { "FN_GLOBAL_BOOL",    "BOOL %s(void)" },
  { "FN_LOCAL_BOOL",     "BOOL %s(int )" },
  { "FN_LOCAL_STRING",   "char *%s(int )" },
  { "FN_GLOBAL_INTEGER", "int %s(void)" },
  { "FN_LOCAL_INTEGER",  "int %s(int )" },
  { "FN_VUSER_STRING",   "char *%s(const user_struct* )" },
  { "FN_LOCAL_CHAR",     "char %s(int )" },
};

static void usage (void)
{
	fprintf(stderr,
		"usage: %s [-f] [-s source_dir] [-d default_header]\n",
		progname);
	exit(1);
}

static char *chompsp (char *str)
{
  char *n = strchr (str, '\n');
  while (n >= str && isspace (*n)) {
    *n-- = '\0';
  }
  return str;
}

static FILE *xfopen (const char *path, char *mode)
{
	FILE *ret = fopen (path, mode);
	if (ret)
		return ret;

	sprintf(errbuf, "%s: cannot open %s", progname, path);
	perror(errbuf);
	exit(1);
}

/*
 * mkdir_p: mkdir() but also create needed parent dirs, like mkdir(1) -p
 * This writes to 'dir' string, so DO NOT pass in a string constant.
 * For arguments, return value and errno values, see mkdir(2)
 */
static int mkdir_p (char *dir, mode_t mode)
{
  int level = 0;
  char *d;

  while (1) {
    struct stat st;
    if (!stat (dir, &st) && (st.st_mode & S_IFDIR)) {
      for (d=dir; level; level--) {
	d += strlen(d);
	*d = '/';
	if (mkdir (dir, mode))
	  return -1;
      }
      return 0;
    }
    d = strrchr (dir, '/');
    if (d) {
      *d = '\0';
      level++;
    } else {
      if (mkdir (dir, mode))
	return -1;
    }
  }
}


/* mkdir -p `dirname $to` && mv $from $to */
static void xrename (const char *from, const char *to)
{
	struct stat st;

	if (!stat(to, &st)) {
		if(unlink(to) < 0)
			goto err;
	}
	else {
		char *p = strrchr(to, '/');
		if (p) {
			char buf[LINELEN];
			memcpy(buf, to, p-to);
			buf[p-to] = 0;
			if (mkdir_p(buf, 0777)) {
				sprintf(errbuf, "%s: cannot make directory %s",
					progname, buf);
				perror(errbuf);
				exit(1);
			}
		}
	}
	if (link (from, to) < 0)
		goto err;
	if (unlink (from) < 0)
		goto err;
	return;
err:
	sprintf(errbuf, "%s: cannot rename %s to %s", progname, from, to);
	perror(errbuf);
	exit(1);
}

static int move_if_change (char *temp, char *file)
{
  struct stat s_t, s_f;

  ZERO_MTIME (s_t); ZERO_MTIME (s_f);
  stat (temp, &s_t);
  stat (file, &s_f);
  if (NEWER (s_t, s_f)) {
    FILE *f_t, *f_f;
    if (s_t.st_size != s_f.st_size)
      goto do_move;
    f_t = xfopen (temp, "r");
    f_f = fopen (file, "r");
    if (!f_f) {
      fclose (f_t);
      goto do_move;
    }
    while (!feof (f_t)) {
      char buf1[BUFLEN], buf2[BUFLEN];
      int n1 = fread (buf1, 1, BUFLEN, f_t);
      int n2 = fread (buf2, 1, BUFLEN, f_f);
      if (n1 != n2 || n1 <= 0 || memcmp (buf1, buf2, n1)) {
	fclose (f_t);
	fclose (f_f);
	goto do_move;
      }
    }
    fclose (f_t);
    fclose (f_f);
  }
  unlink(temp);
  return 0;	/* didn't move file */

 do_move:
  xrename(temp, file);
  return 1;	/* moved file */
}

/*
 * do_xargs: generate argument list from standard input, and return argc.
 * This is to get around arg length limits in various OSes.
 */
static int do_xargs (char ***argv_ret)
{
	char *buf;
	char **argv;
	int argc = 0, bufp = 0, l;
	int buf_max = 16385;
	int argv_max = 1024;

	buf = malloc(buf_max * sizeof(*buf));
	argv = malloc(argv_max * sizeof(*argv));
	if (!buf || !argv) {
    oom:
		fprintf(stderr, "out of memory");
		exit(1);
	}
	while ((l = fread(buf+bufp, 1, buf_max-bufp-1, stdin))) {
		bufp += l;
		if (bufp == buf_max-1) {
			buf_max += 16384;
			buf = realloc(buf, buf_max);
			if (!buf)
				goto oom;
		}
	}
	if (ferror(stdin)) {
		sprintf(errbuf, "%s: error reading input", progname);
		perror(errbuf);
		exit(1);
	}
	buf_max = bufp;
	bufp = 0;
	buf[buf_max] = '\0';

	while (bufp < buf_max) {
		if (argc == argv_max) {
			argv_max += 1024;
			argv = realloc(argv, argv_max);
			if (!argv)
				goto oom;
		}
		while (isspace(buf[bufp]))
			bufp++;
		argv[argc++] = buf+bufp;
		while (buf[bufp] != '\0' && !isspace(buf[bufp]))
			bufp++;
		buf[bufp++] = '\0';
	}
	if (*argv[argc-1] == '\0')
		argc--;
	argv[argc] = NULL;
	*argv_ret = argv;
	return argc;
}




static struct { char *name; FILE *fp; } out_hfile[32];
static int out_hfile_index;

static void get_header_define (char *out, char *in)
{
	char *inp = strrchr(in, '/');
	if (!inp++)
		inp = in;

	*out++ = '_';
	while (*inp) {
		if (isalnum(*inp))
			*out = toupper(*inp);
		else
			*out = '_';
		inp++;
		out++;
	}
	*out++ = '_';
	*out = 0;
}

#define INCLUDE_DIR "include/"
static void get_new_output_header(const char *token, const char *srcfile)
{
	static int rel_start;
	static char tmptok[LINELEN];
	int i, len;
	const char *p;

	if (token && !memcmp(token, OUTPUT_FILE_MAGIC, sizeof(OUTPUT_FILE_MAGIC)-1)) {
		token += sizeof(OUTPUT_FILE_MAGIC)-1;
		while (isspace(*token))
			token++;

		if (! *token)
			token = default_header_name;

		if (!memcmp(token, OUTPUT_FILE_NONE, sizeof(OUTPUT_FILE_NONE)-1)) {
			out_hfile_index = -2;
			return;
		}
	}
	else {
		token = default_header_name;
	}

	len = sizeof(INCLUDE_DIR)-1;
	for (p=token; !isspace(*p) && *p != '*'; p++)
		if (*p == '/')
			len = 0;

	if (!rel_start) {
		rel_start = strlen(source_dir) + 1;
		memcpy(tmptok, source_dir, rel_start-1);
		tmptok[rel_start-1] = '/';
	}
	if (len)
		memcpy(tmptok+rel_start, INCLUDE_DIR, len);
	len += rel_start;
	memcpy(tmptok+len, token, p-token);
	len += p-token;
	tmptok[len] = 0;

	if (!out_hfile[0].name)
		out_hfile[0].name = default_header_name;

	for (i=0; out_hfile[i].name; i++) {
		if (!strcmp(tmptok+rel_start, out_hfile[i].name))
			break;
	}

	if (!out_hfile[i].name)
		out_hfile[i].name = strdup(tmptok+rel_start);

	if (!out_hfile[i].fp) {
		char buf[LINELEN];
		sprintf(buf, HEADER_TEMP, i);
		out_hfile[i].fp = xfopen(buf, "w");

		get_header_define(buf, out_hfile[i].name);

		fprintf(out_hfile[i].fp,
			"#ifndef %s\n"
			"#define %s\n"
			"/* This file is automatically generated "
			"with \"make proto\". DO NOT EDIT */\n"
			"\n", buf, buf);
	}
	if (!strncmp(srcfile, source_dir, rel_start-1))
		srcfile += rel_start;
	fprintf(out_hfile[i].fp,
		"\n/*The following definitions come from  %s  */\n\n", srcfile);

	out_hfile_index = i;
}

static void output_line_to_header (const char *line, const char *infile)
{
	if (out_hfile_index == -2)
		return;
	if (out_hfile_index == -1 ||
	    !out_hfile[out_hfile_index].fp)
		get_new_output_header(0, infile);

	fputs(line, out_hfile[out_hfile_index].fp);
}

static void close_headers (void)
{
	int i;
	printf("%s: updating headers:\n", progname);
	for (i=0; out_hfile[i].fp; i++) {
		int ch;
		char buf1[LINELEN], buf2[LINELEN];
		get_header_define(buf1, out_hfile[i].name);
		fprintf(out_hfile[i].fp, "#endif /* %s */\n", buf1);
		fclose(out_hfile[i].fp);

		sprintf(buf1, HEADER_TEMP, i);
		sprintf(buf2, "%s/%s", source_dir, out_hfile[i].name);
		ch = move_if_change(buf1, buf2);
		printf("  %c %s%s\n",
		       ch?'+':' ', out_hfile[i].name, ch?"":" (unchanged)");
	}
}


static int process_one_uncached (const char *infile, const char *cachefile)
{
  FILE *in, *cache;
  char line[LINELEN];
  int ignore_next = 0, in_header = 0, got_start = 0;

  in = xfopen(infile, "r");
  cache = xfopen(PROTO_TEMP, "w");
  out_hfile_index = -1;

  while (!feof (in) && fgets (line, LINELEN, in)) {
    size_t i;
    char *s;

    chompsp (line);
    if (in_header) {
      int l = strlen (line);
      if (line[l-1] == ')') {
	in_header = 0;
	line[l++] = ';';
      }
      strcpy(line+l, "\n");
      fputs(line, cache);
      output_line_to_header(line, infile);
      continue;
    }

    if (ignore_next) {
      ignore_next = 0;
      continue;
    }

    if (!memcmp (line, OUTPUT_FILE_MAGIC, sizeof(OUTPUT_FILE_MAGIC)-1)) {
      fputs(line, cache); putc('\n', cache);
      get_new_output_header(line, infile);
      continue;
    }

    if (!isalpha (line[0]) || strchr (line, ';'))
      continue;

    for (i=0; i < STRUCT_SIZE (ignore_start_tokens); i++) {
      int l;
      s = ignore_start_tokens[i];
      l = strlen (s);
      if (!strncmp (line, s, l)) {
	if (! line[l]) {
	  ignore_next = 1;
	  break;
	} else if (line[l] == '*' || isspace (line[l]))
	  break;
      }
    }
    if (i < STRUCT_SIZE (ignore_start_tokens))	/* we hit 'break' above */
      continue;

    got_start = 0;
    for (i=0; i < STRUCT_SIZE (start_tokens); i++) {
      int l;
      s = start_tokens[i];
      l = strlen (s);
      if (!strncmp (line, s, l) &&
	  (isspace (line[l]) || line[l]=='*')) {
	got_start = 1;
	break;
      }
    }
    if (! got_start) {
      /* loadparm.c wizardry */
      for (i=0; i < STRUCT_SIZE (macro_tokens); i++) {
	int l;
	char *e;
	s = macro_tokens[i].token;
	l = strlen (s);
	if (!strncmp (line, s, l) && line[l] == '(' &&
	    (e = strchr (line+l, ','))) {
	  char buf[LINELEN];
	  int len;
	  *e = '\0';
	  len = sprintf(buf, macro_tokens[i].fmt, line+l+1);
	  strcpy(buf+len, ";\n");
	  fputs(buf, cache);
	  output_line_to_header(buf, infile);
	  break;
	}
      }
      continue;
    }

    s = strchr (line, '(');
    if (s) {
      char buf[LINELEN];
      int l = s - line + strlen (s);
      memcpy(buf, line, l);
      if (line[l-1] == ')')
	buf[l++] = ';';
      else
	in_header = 1;
      buf[l++] = '\n';
      buf[l] = 0;
      fputs(buf, cache);
      output_line_to_header(buf, infile);
      continue;
    }
  }
  fclose(in);
  fclose(cache);
  xrename(PROTO_TEMP, cachefile);
  return 0;
}

static int process_one_cached (const char *infile, const char *cachefile)
{
	FILE *fp;
	char line[LINELEN];

	out_hfile_index = -1;
	fp = xfopen(cachefile, "r");
	while (!feof (fp) && fgets (line, LINELEN, fp)) {
		if (!memcmp(line, OUTPUT_FILE_MAGIC, sizeof(OUTPUT_FILE_MAGIC)-1)) {
			get_new_output_header(line, infile);
			continue;
		}
		output_line_to_header(line, infile);
	}
	fclose(fp);
	return 0;
}

static int process_files (int nfiles, char *file[])
{
  int i, j;
  char pname[LINELEN], fname[LINELEN];
  int plen, flen;

  sprintf (pname, "%s/", proto_dir);
  sprintf (fname, "%s/", source_dir);
  plen = strlen (pname);
  flen = strlen (fname);

  /* Have to 'sort -u' them.  Break out the bubble sort. (: */
  for (i=0; i<nfiles; i++) {
    for (j=i+1; j<nfiles; j++) {
      char *tmp;
      int k = strcmp (file[i], file[j]);
      if (k == 0)		/* duplicate -- mark it */
	*file[i] = '\0';
      else if (k > 0) {
	tmp = file[i]; file[i] = file[j]; file[j] = tmp;
      }
    }
  }
  /* Remove duplicates marked above. */
  for (j=i=0; j<nfiles; i++,j++) {
    while (!*file[j] && j < nfiles)
      j++;
    if (i < j && j < nfiles)
      file[i] = file[j];
  }
  nfiles = i;

  for (i=0; i<nfiles; i++) {
    struct stat st_f, st_p;
    int l = strlen(file[i]);
    char *input = fname;

    if (file[i][l-2] == '.' && file[i][l-1] == 'o')
      file[i][l-1] = 'c';

    memcpy(pname+plen, file[i], l+1);
    memcpy(fname+flen, file[i], l+1);

    if (stat (pname, &st_p))
      ZERO_MTIME(st_p);

    if (stat (input, &st_f)) {
      /* look again: source file may be in build directory */
      input = file[i];
      if (stat (input, &st_f)) {
	fprintf (stderr, "missing input file: %s\n", input);
	return -1;
      }
    }
    if (force || NEWER (st_f, st_p))
      process_one_uncached(input, pname);
    else
      process_one_cached(input, pname);
  }

  close_headers();
  return 0;
}

int main (int argc, char *argv[])
{
	int c;
	extern char *optarg;
	extern int optind;

	progname = strrchr(argv[0], '/');
	if (!progname++)
		progname = argv[0];

	while ((c = getopt(argc, argv, "d:s:f")) != -1) {
		switch (c) {
		    case 'd':
			default_header_name = optarg; break;
		    case 's':
			source_dir = optarg; break;
		    case 'f':
			force = 1; break;
		    default:
			usage();
		}
	}
	if (argc-optind > 1)
		usage();
	argc = do_xargs(&argv);
	return process_files(argc, argv);
}
