/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Password and authentication handling
   Copyright (C) Andrew Tridgell 1992-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include <pwd.h>
#ifdef HAVE_NETGROUP_H
#include <netgroup.h>
#endif
#include "libsamba.h"
#include "smbd.h"


static BOOL check_user_equiv_line(char *buf, const char *user,
				  const char *remote, BOOL *plus_allowed,
				  BOOL *ret)
{
	BOOL is_group = False;
	int plus = 1;
	char *bp = buf;
	char *file_host;
	char *file_user;

	if (strcmp(buf, "NO_PLUS\n") == 0)
	{
		DEBUG(6, ("check_user_equiv NO_PLUS\n"));
		(*plus_allowed) = 0;
		return False;
	}

	if (buf[0] == '+')
	{
		bp++;
		if (*bp == '\n' && (*plus_allowed))
		{
			/* a bare plus means everbody allowed */
			DEBUG(6, ("check_user_equiv everybody allowed\n"));
			*ret = True;
			return True;
		}
	}
	else if (buf[0] == '-')
	{
		bp++;
		plus = 0;
	}
	if (*bp == '@')
	{
		is_group = True;
		bp++;
	}

	file_host = strtok(bp, " \t\n");
	file_user = strtok(NULL, " \t\n");
	DEBUG(7, ("check_user_equiv %s %s\n",
		  file_host ? file_host : "(null)",
		  file_user ? file_user : "(null)"));

	if (file_host && *file_host)
	{
		BOOL host_ok = False;
#if defined(HAVE_NETGROUP) && defined(HAVE_YP_GET_DEFAULT_DOMAIN)
		if (is_group)
		{
			static char *mydomain = NULL;
			if (!mydomain)
				yp_get_default_domain(&mydomain);
			if (mydomain && innetgr(file_host, remote, user,
						mydomain)) host_ok = True;
		}
#else
		if (is_group)
		{
			DEBUG(1, ("Netgroups not configured\n"));
			return False;
		}
#endif
		/* is it this host */
		/* the fact that remote has come from a call of
		 * gethostbyaddr means that it may have the fully
		 * qualified domain name so we could look up the
		 * file version to get it into a canonical form,
		 * but I would rather just type it in full in the
		 * equiv file */
		if (!host_ok && !is_group && strequal(remote, file_host))
		{
			host_ok = True;
		}
		if (!host_ok)
		{
			return False;
		}
		/* is it this user */
		if (file_user == 0 || strequal(user, file_user))
		{
			DEBUG(5, ("check_user_equiv matched %s%s %s\n",
				  (plus ? "+" : "-"), file_host,
				  (file_user ? file_user : "")));
			*ret = (plus ? True : False);
			return True;
		}
	}

	return False;
}

/****************************************************************************
read the a hosts.equiv or .rhosts file and check if it
allows this user from this machine
****************************************************************************/
static BOOL check_user_equiv(char *user, char *remote, char *equiv_file)
{
	pstring buf;
	int plus_allowed = 1;
	FILE *fp = sys_fopen(equiv_file, "r");
	DEBUG(5, ("check_user_equiv %s %s %s\n", user, remote, equiv_file));
	if (!fp)
		return False;
	while (fgets(buf, sizeof(buf), fp))
	{
		trim_string(buf, " ", " ");

		if (buf[0] != '#' && buf[0] != '\n')
		{
			BOOL ret;
			if (check_user_equiv_line(buf, user, remote,
						  &plus_allowed, &ret))
			{
				fclose(fp);
				return ret;
			}
		}
	}
	fclose(fp);
	return False;
}


/****************************************************************************
check for a possible hosts equiv or rhosts entry for the user
****************************************************************************/
BOOL check_hosts_equiv(char *user)
{
	char *fname = NULL;
	pstring rhostsfile;
	const struct passwd *pass = Get_Pwnam(user, True);

	if (!pass)
		return(False);

	fname = lp_hosts_equiv();

	/* note: don't allow hosts.equiv on root */
	if (fname && *fname && (pass->pw_uid != 0)) {
		if (check_user_equiv(user, client_name(), fname))
			return (True);
	}

	if (lp_use_rhosts())
	{
		char *home = get_user_home_dir(user);
		if (home)
		{
			slprintf(rhostsfile, sizeof(rhostsfile) - 1,
				 "%s/.rhosts", home);
			if (check_user_equiv(user, client_name(), rhostsfile))
				return (True);
		}
	}

	return False;
}
